#Opening file
from pathlib import Path

# Organizing data
from dataclasses import dataclass
import pandas as pd

# Argument parsing, regular expressions
import argparse, re 

# Data Visualization libraries
import seaborn as sns
import matplotlib as mplib
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
from matplotlib.colors import LinearSegmentedColormap

# Custom legend
from matplotlib.lines import Line2D

# Opening images
from PIL import Image

#Saving to PDF
from matplotlib.backends.backend_pdf import PdfPages
# https://stackoverflow.com/questions/11328958/save-multiple-plots-in-a-single-pdf-file?rq=3
# https://stackoverflow.com/questions/21364405/saving-plots-to-pdf-files-using-matplotlib

#https://www.sonofacorner.com/beautiful-tables/

# Copy objects
import copy

@dataclass
class overviewTable:
    bus_width_bits: int
    timeStep: float
    voltageStep: float
    tool_version: str
    num_samples: int
    swizzleList: list[int]
    ranks:  int
    channels: int

@dataclass
class writeEyeTables:
    dram_vrefs: list[float]
    maxVoltageHeight: list[float] #Per DQ
    maxTimingWidth: list[float] # Per DQ
    totalPassingPointsDQ: list[int]
    worstVoltageHeight: float
    worstTimingWidth: float
    smallestPassingPoints: int

@dataclass 
class readEyeTables:
    soc_vrefs: list[float]
    maxVoltageHeight: list[float]
    maxTimingWidth: list[float]
    totalPassingPointsDQ: list[int]
    worstVoltageHeight: float
    worstTimingWidth: float
    smallestPassingPoints: int



@dataclass
class eyeDataPoint:
    fail_cnt:   int  
    vref:       float   
    delay:      float     

@dataclass
class eyeDataTrainedPoint:
    type:           str # Rise/Fall/Write
    physicalCS:     int #
    logicalCS:      int #
    vref:           float   
    delay:          float   
       

@dataclass
class eyeDiagram:
    dq:                 int                 # Contains the data bit for this eye diagram
    edge:               str                 # Contains the edge for the bit
    #type:               str                 # Contains type, if read or write
    eyeData:            list[eyeDataPoint]  # Long form vector of fails for seaborn
    trainData:          list[eyeDataTrainedPoint]


@dataclass 
class allEyeDiagrams:
    type:       str
    diagrams:   list[eyeDiagram]
    max_seen:   int
    
global summaryTable
global writeTable
global readTable
global readMarginPlot # Class of allEyeDiagrams
global writeMarginPlot # Class of allEyeDiagrams
global eye_type
global physical_CS
global logical_CS
global bus_width
global num_vref
global num_rddly
global subPlotAspectRatio
global vref_limit
global rddly_limit
global wrdly_limit
global wrdly_min

def convertMR14toVREF(MR14):
    return( 100 + ((MR14 >> 6)&1)*110 + 4*(MR14 & (63) ))

def convertSoCVREFtoVREF(socVREF):
    if ((socVREF >> 8) & 0x1) == 1:
        if debug:
            print("Range 0 detected")
        return (54000 + (socVREF & 127)*2638)
    else:
        if debug:
            print("Range 1 detected")
        return (119000 + (socVREF & 127)*2874)
    

def convertAnyVREFtoVoltage(vref, eye_type):
    if eye_type == "READ":
        return convertReadVREFtoVoltage(vref)
    if eye_type == "WRITE":
        return convertWriteVREFtoVoltage(vref)
    
def convertReadVREFtoVoltage(vref):
    #Uses VDDQ as VREF, which is 1.1V nominally
    return (vref/(10000*100))*1100

def convertWriteVREFtoVoltage(vref):
    # JEDEC defines Vref to be VDDQ, which is 1.1V nominally
    return (vref/(10*100))*1100

def convertDeltaDelaytoUnitInterval(delay):
    # Delta delay is fraction of a cycle over 512
    # One unit interval  = 0.5 clock cycles
    # Inversely, one clock cycle = 2 Unit Intervals
        # To convert delta delay to cycle, divide by 512
        # Then multiply by 2 to convert to unit interval
    return (delay/256.0)





def parseLogfile(filename):
    global debug
    global bus_width
    global readMarginPlot
    global writeMarginPlot
    global num_vref 
    global num_rddly 
    global vref_limit
    global wrdly_limit
    global rddly_limit
    global wrdly_min
    global eye_type
    global physical_CS
    global logical_CS
    global summaryTable 
    global writeTable
    global readTable
    rdly_min = 0

    if debug: 
        print(filename)

    max_seen = 0
    num_vref = 0
    num_rddly = 0
    try:   
        csv_handle= open(filename, "r")
    except IOError:
        print("File <" + filename + "> not found!")

    
    bus_width=0
    physical_CS = 0
    byte_lanes=0
    soc_type = "NONE"
    eye_type = "NONE"

    tool_major_version = 1
    tool_minor_version = 0
    tool_patch_version = 0
    tool_version_str = ""
    # Find start of relevant log

    
    for line in csv_handle:
        line = line.strip()
        if "DDRSS \"Teye\" Tool" in line:
            #print("Found start")
            continue
            #break   
        versionRegex = re.compile('Version (\d+)\.(\d+)\.(\d+)')
        m = versionRegex.search(line)
        if m is not None:
            tool_major_version = int(m.group(1))
            tool_minor_version = int(m.group(2))
            tool_patch_version = int(m.group(3))
            tool_version_str = m.group(1) + "."+ m.group(2) +"." + m.group(3)
            print("Tool version is " + m.group(1) + "."+ m.group(2) +"." + m.group(3))

            break
        if "TEYE_TOOL_START" in line:
            #continue
            print("Found start")
            break   

    # Find SoC
    for line in csv_handle:
        line = line.strip()
        # SOC Type=SOC_AM62A
        socRegex = re.compile('SOC Type=(\w+)$')
        m = socRegex.search(line)
        if m is not None:
            soc_type = m.group(1)
            break

    if debug: 
        print("SoC found is " + soc_type)

    # Find number of DQ bits       
    for line in csv_handle:
        line = line.strip()
        # DETECTED: Bus width: 32 bits wide
        busWidthRegex = re.compile('DETECTED: Bus width: (\d+) bits wide$')
        m = busWidthRegex.search(line)
        if m is not None:
            bus_width = int(m.group(1))
            break
        
    # Error out if bus width not found    
    if(bus_width == 0):
        print("ERROR: Bus Width not found!")
        return
    
    if debug: 
        print("Bus width found is {}".format(bus_width))

    # Decode bus width to number of byte lanes
    byte_lanes = int(bus_width/8)

    # Find number of logical chip selects/ranks
    # Used to determine how mnay training values are displayed on plots
    for line in csv_handle:
        line = line.strip()
        rankRegex = re.compile('DETECTED: Ranks: (\d)$')
        m = rankRegex.search(line)
        if m is not None:
            logical_CS = int(m.group(1))
            break

    # Find number of physical chip selects
    for line in csv_handle:
        line = line.strip()
        physicalCSRegex = re.compile('DETECTED: Physical chip selects: (\d)$')
        m = physicalCSRegex.search(line)
        if m is not None:
            physical_CS = int(m.group(1))
            break
    
    # Default physical to logical CS mapping
    # Rank 0 = Phys CS 0, Phys CS 2
    # Rank 1 = Phys CS 1, Phys CS 3
    phys2logCS = {0:0, 1:1, 2:0, 3:1}
    swizzleList = []

    # for i in range(0, byte_lanes):
    #     swizzleList.append(list())
    #     for j in range(0,9):  
    #         swizzleList[i].append(j)

    #DQ bits 0 to bus_width-1 (for example, 0 to 31) for 32 bit bus
    for i in range(0, bus_width):
        swizzleList.append(i)
    #DM bits bus_width to bus_width+byte_lanes - 1, (For example, 32 to 35 to map DM0-3 for a 32 bit bus)
    for j in range(bus_width,bus_width+byte_lanes):  
        swizzleList.append(j)   

    if (tool_major_version>=1 and tool_minor_version >=8) or (tool_patch_version != 0):
        print("Tool version has swizzling, decoding now!")
        matched_bits=0
        for line in csv_handle:
            line = line.strip()
     
            swizzlingRegex = re.compile('PAD_D(Q|M)(\d+) = DRAM_D(Q|M)(\d+)')
            #print(line)
            m = swizzlingRegex.search(line)
            if m is not None:
                matched_bits+=1
                #print(line)
                if m.group(3) == "Q":
                    # slice = int(int(m.group(2))/8)
                    # bit = int(int(m.group(2))%8)
                    dq_bit = int(m.group(4))
                    if m.group(1) == "Q":
                        #swizzleList[slice][bit] = int(int(m.group(4))%8)
                        swizzleList[dq_bit] = int(m.group(2))
                    if m.group(1) == "M":
                        #swizzleList[slice][bit] = 8
                        swizzleList[dq_bit] = bus_width + int(m.group(2))
                if m.group(3) == "M":
                    #slice = int(m.group(2))
                    dm_bit = int(m.group(4))
                    if m.group(1) == "Q":
                        #swizzleList[slice][8] = int(int(m.group(4))%8)
                        swizzleList[dm_bit + bus_width] = int(m.group(2))
                    if m.group(1) == "M":
                        #swizzleList[slice][8] = 8
                        swizzleList[dm_bit + bus_width] = bus_width + int(m.group(2))
                        
            if(matched_bits == (byte_lanes*9)):
                break

    else:
        print("Tool version used was an early dev build, does not print out the swizzling information or has incorrect information! Default mapping will be used!")
        print("To add swizzling information, manually edit log and then change patch version (ex. 1.7.0 -> 1.7.1 )")
    
    if debug:
        # for i in range(0, byte_lanes):
        #     for j in range(0,8):  
        #         if(swizzleList[i][j] == 8):
        #             print("DQ" + str(j+ 8*i) +" = DM" + str(i)  )
        #         else:
        #             print("DQ" + str(j+ 8*i) +" = DQ" + str(8*i + swizzleList[i][j]) )
        #     if(swizzleList[i][8] == 8):
        #         print("DM" + str(i) +" = DM" + str(i)  )
        #     else:
        #         print("DM" + str(i) +" = DQ" + str(8*i + swizzleList[i][8]) )
        for i in range(0, bus_width):
            if(swizzleList[i] >= bus_width):
                print("DQ" + str(i) +" = DM" + str(swizzleList[i]-bus_width)  )
            else:
                print("DQ" + str(i) +" = DQ" + str(swizzleList[i]) )
        for i in range(0, byte_lanes):
            if(swizzleList[bus_width + i] >= bus_width):
                print("DM" + str(i) +" = DM" + str(swizzleList[i]-bus_width)  )
            else:
                print("DM" + str(i) +" = DQ" + str(swizzleList[i]) )


    # Start to fill in overview table
    summaryTable = overviewTable(bus_width, 0, 0, tool_version_str, 0, swizzleList, logical_CS, bus_width/16) 

    # Find what eye(s) are in this log
    # Can be one (READ or WRITE) or both
    for line in csv_handle:
        line = line.strip()
        #Teye_eye_type=
        eyeTypeRegex= re.compile('Teye_eye_type=(\w+)$')
        m = eyeTypeRegex.search(line)
        if m is not None:
            eye_type = m.group(1)
            break
    if debug: 
        print("Eye type found is " + eye_type)

    # Grab all DRAM VREF settings, unique value per FSP and physical CS
    # Ex: FSP 2 will have a MR14 value for every physical CS line that goes to the DRAM, since a DRAM can have multiple dies of DRAM/MR14
    mr14 =[0]*physical_CS
    mr_matches = 0
    for line in csv_handle:
        line = line.strip()
        mr14Regex = re.compile('pi_mr14_f2_cs(\d).+= (\w+)$')
        m = mr14Regex.search(line)
        if m is not None:
            mr_matches+=1
            #print('Matched line: ' + line)
            cs = int(m.group(1))
            mr14[cs] = convertMR14toVREF( int(m.group(2), 16) )
            # print('CS = ' + str(cs) )
            # print('MRVAL = ' + str( mr14[cs] ))
        if mr_matches == physical_CS:
            break


    # Fill in DRAM VREFs

    writeTable = writeEyeTables(mr14, [], [], [], 0, 0, 0 )

    # Grab all SoC VREF settings, unique value per FSP and bytelane/slice
    # Ex: FSP 2 will have a SoC VREF setting per byte lane/slice
    soc_vref =[0]*byte_lanes
    soc_vref_matches = 0
    for line in csv_handle:
        line = line.strip()

        # Case for old incorrect syntax
        # This line previously attributed this to CS, whereas this is really per byte lane
        soc_vrefRegex = re.compile('phy_vref_ctrl_dq_vrefsel_f2_cs(\d).+= (\w+)$')
        m = soc_vrefRegex.search(line)
        if m is not None:
            soc_vref_matches+=1
            #print('Matched line: ' + line)
            cs = int(m.group(1))
            soc_vref[cs] = convertSoCVREFtoVREF( int(m.group(2), 16) )
            if debug:
                print("Old syntax")
                print(str(soc_vref))
            # print('CS = ' + str(cs) )
            # print('SOCVREF = ' + str( soc_vref[cs] ))
        if soc_vref_matches == byte_lanes:
            if debug:
                print(line)
            break

        # Case for updated syntax
        soc_vrefRegex = re.compile('phy_vref_ctrl_dq_vrefsel_f2_slice(\d).+= (\w+)$')
        m = soc_vrefRegex.search(line)
        if m is not None:
            soc_vref_matches+=1
            #print('Matched line: ' + line)
            cs = int(m.group(1))
            soc_vref[cs] = convertSoCVREFtoVREF( int(m.group(2), 16) )
            if debug:
                print("Updated syntax")
                print(str(soc_vref))
            # print('CS = ' + str(cs) )
            # print('SOCVREF = ' + str( soc_vref[cs] ))
        if soc_vref_matches == byte_lanes:
            if debug:
                print(line)
            break

    readTable = readEyeTables(soc_vref, [], [], [], 0, 0, 0 )



    # Grab all trained delay values, unique values per FSP, DQ, and logical CS (rank)
    # Ex: FSP2 will have a unique set of read delay rise+fall and write delay per DQ for CS0 (rank 0, can possibly be two physical CS), and then another unique set for CS1 (rank 1, can possibly be two physical CS)
    write_delay = [0]*bus_width*logical_CS
    read_delay_rise = [0]*bus_width*logical_CS
    read_delay_fall = [0]*bus_width*logical_CS
    write_delay_dm = [0]*byte_lanes*logical_CS
    read_delay_rise_dm = [0]*byte_lanes*logical_CS
    read_delay_fall_dm = [0]*byte_lanes*logical_CS
    delay_matches = 0
    for line in csv_handle:
        line = line.strip()
        write_delay_regex = re.compile('write_delay, f(\d)_cs(\d)_slice(\d)_dq(\d).+= (\w+)$')
        read_delay_fall_regex = re.compile('read_delay_fall, f(\d)_cs(\d)_slice(\d)_dq(\d).+= (\w+)$')
        read_delay_rise_regex = re.compile('read_delay_rise, f(\d)_cs(\d)_slice(\d)_dq(\d).+= (\w+)$')

        write_delay_dm_regex = re.compile('write_delay_dm, f(\d)_cs(\d)_slice(\d)\s+= (\w+)$')
        read_delay_fall_dm_regex = re.compile('read_delay_dm_fall, f(\d)_cs(\d)_slice(\d)\s+= (\w+)$')
        read_delay_rise_dm_regex = re.compile('read_delay_dm_rise, f(\d)_cs(\d)_slice(\d)\s+= (\w+)$')

        m = write_delay_regex.search(line)
        if m is not None:
            if int(m.group(1)) == 2:
                delay_matches+=1
                cs = int(m.group(2))
                slice = int(m.group(3))
                dq = int(m.group(4))
                write_delay[ cs*bus_width +  8*slice + dq   ] = int(m.group(5), 16)
                #print(line)

        m = read_delay_fall_regex.search(line)
        if m is not None:
            if int(m.group(1)) == 2:
                delay_matches+=1
                cs = int(m.group(2))
                slice = int(m.group(3))
                dq = int(m.group(4))
                read_delay_fall[ cs*bus_width +  8*slice + dq   ] = int(m.group(5), 16)
                #print(line)

        m = read_delay_rise_regex.search(line)
        if m is not None:
            if int(m.group(1)) == 2:
                delay_matches+=1
                cs = int(m.group(2))
                slice = int(m.group(3))
                dq = int(m.group(4))
                read_delay_rise[ cs*bus_width +  8*slice + dq   ] = int(m.group(5), 16)
                #print(line)

        m = write_delay_dm_regex.search(line)
        if m is not None:
            if int(m.group(1)) == 2:
                delay_matches+=1
                cs = int(m.group(2))
                slice = int(m.group(3))   
                write_delay_dm[ cs*byte_lanes +  slice   ] = int(m.group(4), 16)
                #print(line)

        m = read_delay_fall_dm_regex.search(line)
        if m is not None:
            if int(m.group(1)) == 2:
                delay_matches+=1
                cs = int(m.group(2))
                slice = int(m.group(3))
                read_delay_fall_dm[ cs*byte_lanes +  slice    ] = int(m.group(4), 16)
                #print(line)

        m = read_delay_rise_dm_regex.search(line)
        if m is not None:
            if int(m.group(1)) == 2:
                delay_matches+=1
                cs = int(m.group(2))
                slice = int(m.group(3))
                read_delay_rise_dm[  cs*byte_lanes +  slice    ] = int(m.group(4), 16)
                #print(line)
                
        # No matter what type of eye (read or write), we capture 3 training delay values for every DQ and DM bit
        # Each of these delay values vary per FSP and DQ/DM and CS
        # This for loop is for a single FSP, so just need to account for CS, DQ/DM bits
        if delay_matches == (bus_width*logical_CS*3) + ((bus_width/8)*logical_CS*3) :
            if debug:
                print(line)
            break

 

    # Populate initial eye diagram structures based off eye type
    # If read, populate readMarginPlot with empty values
    # If write, populate writeMarginPlot with empty values
    # If both, do both above
    # Each margin plot will have two diagrams per bit, one for rise and one for fall
    # Number of empty values is based on how many DQ bits exists
    
    eyeDiagrams = []
    numEyes=0
    
    if "READ" in eye_type:
        numEyes+=1
        
        readMarginPlot = allEyeDiagrams("READ", [], 0)
        for i in range(0, bus_width):
            slice = int(i/8)
            bit = int(i%8)
            #print("Slice = " + str(slice) + ", bit = " + str(bit))
            # Populate an initial eye diagram with an empty array for the eyedata points and training values
            readMarginPlot.diagrams.append(eyeDiagram(i, "rise", [], []))
            readMarginPlot.diagrams.append(eyeDiagram(i, "fall", [], []))

            # Each DQ will have 2 training points per logical CS
            # One point for rise, one point for fall
            for j in range(0, logical_CS):
                
                # if(swizzleList[slice][bit] == 8 ): #DM bit
                #     trained_rise = read_delay_rise_dm[ j*byte_lanes + slice ]
                #     trained_fall = read_delay_fall_dm[ j*byte_lanes + slice ]
                # else:
                #     trained_rise = read_delay_rise[ j*bus_width + 8*slice + swizzleList[slice][bit] ]
                #     trained_fall = read_delay_fall[ j*bus_width + 8*slice + swizzleList[slice][bit] ]
                if(swizzleList[i] >= bus_width ): #DM bit
                    trained_rise = read_delay_rise_dm[ j*byte_lanes + (swizzleList[i] -bus_width) ]
                    trained_fall = read_delay_fall_dm[ j*byte_lanes + (swizzleList[i] -bus_width) ]
                else:
                    trained_rise = read_delay_rise[ j*bus_width + swizzleList[i] ]
                    trained_fall = read_delay_fall[ j*bus_width + swizzleList[i] ]
                # Populate physical CS with -1 for N/A
                readMarginPlot.diagrams[(i*2)    ].trainData.append(eyeDataTrainedPoint("rise", -1, j, soc_vref[slice], trained_rise))
                readMarginPlot.diagrams[(i*2) + 1].trainData.append(eyeDataTrainedPoint("fall", -1, j, soc_vref[slice], trained_fall))

    if "WRITE" in eye_type:
        numEyes+=1
        writeMarginPlot = allEyeDiagrams("WRITE", [], 0)
        for i in range(0, bus_width):
            slice = int(i/8)
            bit = int(i%8)
            
            # Populate an initial eye diagram with an empty array for the eyedata points and training values
            writeMarginPlot.diagrams.append(eyeDiagram(i, "rise", [], []))
            writeMarginPlot.diagrams.append(eyeDiagram(i, "fall", [], []))

            # Each DQ will have 1 training point per physical CS
            # No rise/fall uniqueness
            for j in range(0, physical_CS):
                # if(swizzleList[slice][bit] == 8 ):
                #     trained_delay = write_delay_dm[ phys2logCS[j]*byte_lanes+  slice ]
                # else:
                #     trained_delay = write_delay[ phys2logCS[j]*bus_width + 8*slice + swizzleList[slice][bit] ]
                if(swizzleList[i] >= bus_width ):
                    trained_delay = write_delay_dm[ phys2logCS[j]*byte_lanes+  (swizzleList[i] -bus_width) ]
                else:
                    trained_delay = write_delay[ phys2logCS[j]*bus_width + swizzleList[i] ]
                
                writeMarginPlot.diagrams[(i*2)    ].trainData.append(eyeDataTrainedPoint("rise", j, phys2logCS[j], mr14[j], trained_delay))
                writeMarginPlot.diagrams[(i*2) + 1].trainData.append(eyeDataTrainedPoint("fall", j, phys2logCS[j], mr14[j], trained_delay))


    if (tool_major_version>=1 and tool_minor_version >=8) :
        for line in csv_handle:
            line = line.strip()
            samples_regex = re.compile('Total transfer size is (\d+) Bytes$')
            m = samples_regex.search(line)
            if m is not None:
                if debug:
                    print("Found samples!")
                summaryTable.num_samples = str(int((int(m.group(1)))/int(byte_lanes))) + " samples per bit"
                break
    else:
        print("Tool version used was an early dev build, does not print out number of samples!")
        summaryTable.num_samples = "Pre-release: " + str(int(2097152/(int(byte_lanes)))) + " samples per bit"


    # Generic regex for any number at the end of a line
    eolNumberRegex = re.compile('(\d+)$')
    for i in range(0,numEyes):
        vref = -1
        dly = 0
        edge = "none"
        dqs_num = 0
        dly_min =0
        dly_limit =0
        activeEye = None 
        max_seen = 0
        # Support legacy log kinda
        #If
        if "READ_WRITE" in eye_type:
            #print("Detected read and write in one log")
            for line in csv_handle:
                line = line.strip()
                activeEyeRegex = re.compile('^(\w+)_EYE_DATA_COLLECTION_START')
                m = activeEyeRegex.search(line)
                if m is not None:     
                    activeEye = m.group(1)
                    break
        else:
            activeEye = eye_type
            for line in csv_handle:
                line = line.strip()
                if "vref=" in line:
                    m = eolNumberRegex.search(line)
                    vref = int(m.group(1))
                    break

        print("Processing " + activeEye + " eye log")
        for line in csv_handle:
            line = line.strip()
            if "vref=" in line:
                m = eolNumberRegex.search(line)
                vref = int(m.group(1))

                #if activeEye == "READ":
                    #vref = convertReadVREFtoVoltage(vref)
                #if activeEye == "WRITE":
                    #vref = convertWriteVREFtoVoltage(vref)

                #print(str(vref))
                continue
            if "rddly=" in line:
                m = eolNumberRegex.search(line)
                dly = int(m.group(1))
                dly_min = rdly_min
                dly_limit = rddly_limit
                continue
            if "wrdly=" in line:
                m = eolNumberRegex.search(line)
                dly = int(m.group(1))
                dly_min = wrdly_min
                dly_limit = wrdly_limit
                continue
            if "dqs_edge=" in line:
                m = eolNumberRegex.search(line)
                dqs_num = int(m.group(1))
                if(dqs_num == 0):
                    edge = "rise"
                else:
                    edge = "fall"
                continue
            if "DATA_COLLECTION_DONE" in line:
                if debug: 
                    print("found done")
                #end if debug
                break
            if "temperature=" in line:
                continue
            elif "MR14" in line:
                continue
            elif "ERROR" in line:
                continue
            elif "mode+en+sel" in line:
                continue
            # elif len(line.split()) == 0: #Check for empty line
            #     continue
            else:
                dq_bit = 0
                for fail_string in line.split(','):
                    fail_num = int(fail_string)
                    if(fail_num > max_seen):
                        max_seen = fail_num
                        if debug: 
                            print("Max seen is " + fail_string)
                        #end if debug
                    # end if

                    # if debug: 
                    #     if( (dq_bit%2 == 0 ) & (dqs_num==0) & (dly == 160) & (vref ==40) ):
                    #         fail_num = 16000
                    #         print("Inserting fake fail point for even rising dq bits\n")

                    #     if( (dq_bit%2 == 1 ) & (dqs_num==1) & (dly == 160) & (vref ==40) ):
                    #         fail_num = 16000
                    #         print("Inserting fake fail point for odd falling dq bits\n")
                    # end if debug 

                    if vref_limit != None and vref > int(vref_limit):
                        #print("Found higher vref than max at ")
                        continue
                    if dly_limit != None and dly > int(dly_limit):
                        #print("Found higher dly than max at ")
                        continue
                    if dly_min != None and dly < int(dly_min):
                        #print("Found lower dly than min at ")
                        continue

                    if "READ" in activeEye:
                        readMarginPlot.max_seen = max_seen
                        readMarginPlot.diagrams[dq_bit*2 + dqs_num].eyeData.append(eyeDataPoint(fail_num,vref,dly))
                    if "WRITE" in activeEye:
                        writeMarginPlot.max_seen = max_seen
                        writeMarginPlot.diagrams[dq_bit*2 + dqs_num].eyeData.append(eyeDataPoint(fail_num,vref,dly))
                    #eyeDiagrams[dq_bit*2 + dqs_num].eyeData.append(eyeDataPoint(fail_num,vref,dly))
                    dq_bit += 1
            
    # eyeVectors = pd.DataFrame(eyeDiagrams)
    # print(eyeVectors)



def construct_pdf_report(pdf_handle, fullMarginPlot : allEyeDiagrams, filename_no_ext, color_plot) :
    global bus_width
    global eyeDiagrams 
    global subPlotAspectRatio
    global vref_limit
    global wrdly_min
    global wrdly_limit
    global physical_CS
    global logical_CS
   # DONE: Custom colormap / palette for eye diagrams
    # https://stackoverflow.com/questions/42656585/barplot-colored-according-a-colormap/42665818#42665818
    # https://matplotlib.org/2.0.2/examples/pylab_examples/custom_cmap.html
    # https://matplotlib.org/stable/gallery/color/custom_cmap.html
    # >>> Look at this one https://matplotlib.org/stable/tutorials/colors/colormap-manipulation.html

    eyeVectors_rise = []
    eyeVectors_fall = []
    eyeVectors_sum = []


    max_seen = fullMarginPlot.max_seen
    eye_type = fullMarginPlot.type
    eyeDiagrams = copy.deepcopy(fullMarginPlot.diagrams)


    # Test out aspect ratio of plots
    df_test = pd.DataFrame(eyeDiagrams[0].eyeData)
    subPlotAspectRatio = df_test.vref.unique().size/df_test.delay.unique().size
    print("Aspect Ratio of each subplot is: " + str(subPlotAspectRatio) )
    if(subPlotAspectRatio<0.8 or subPlotAspectRatio > 2):
        print("Skewed dataset, results may not look good...")


    # TODO: Revisit support for colored eye
    #Use this for each individual rise/fall eye
    colors1 =  [ (0, "white"), (1/max_seen, "blue"), (1, "red")]
    cmap1 = LinearSegmentedColormap.from_list("test", colors1, N=max_seen)

    #Use this for the grand total eye
    colors2 =  [ (0, "white"), (1/(2*max_seen), "blue"), (1, "red")]
    cmap2 = LinearSegmentedColormap.from_list("test", colors2, N=2*max_seen)  

    #Use this for black and white
    # colors3 = [ (0, "white"), (1, "black")]
    # cmap3 = LinearSegmentedColormap.from_list("test", colors3, N=1)
    # Colorized plot
    colors_palette =  [ (0, "white"), (1/max_seen, "blue"), (1, "red")]
    cmap_palette = LinearSegmentedColormap.from_list("test", colors_palette, N=max_seen)


    # Create a palette and list of markers colors based on type of eye and how many logical/physical CS present
    #                         Pass     Fail     CS0      CS1     CS2     CS3            
    full_training_palette =                   ["blue", "green", "red", "purple"]
    # This is the bare minimum, single rank single channel = 1 CS (16 bit, 1 channel is defined as 16 bit in JEDEC spec)
    # Next permutation would be single rank dual channel = 1 logical CS, two physical CS per logical CS (32 bit, spread over 2 16 bit devices)
    # Not likely to be used, but dual rank single channel = 2 logical CS, two physical CS but broken down to one physical CS per logical CS (16 bit x 2 devices that have 16 bit)
    # Most likely to be used when possible, dual rank dual channel =  2 logical CS, 4 physical CS, broken donw to 2 physical CS per logical CS (32 bit x 2 devices that have 32 bit)

    # In this case we will change the color palette by eye type to reflect logical and physical CS
    
    # TODO: Create a sizing for the dots based on math of dpi, pixels, etc.
    #                        Pass   Fail

    # legend_elements = [ Line2D([0], [0], marker = 's', color = 'w', markerfacecolor='black', label = 'Failing point'),
    #                 Line2D([0], [0], marker = '^', color = 'w', markerfacecolor='green', label = 'Rise'),
    #                 Line2D([0], [0], marker = 'v', color = 'w', markerfacecolor='blue', label = 'Fall'),
    #                 Line2D([0], [0], marker = 'X', color = 'w', markerfacecolor='red', label = 'Other')
    #                 ]
        
    legend_elements = [ Line2D([0], [0], marker = 's', color = 'w', markeredgecolor='black', markeredgewidth=0.1, markerfacecolor='white', label = 'Passing point'),
                        Line2D([0], [0], marker = 's', color = 'w', markerfacecolor='black', label = 'Failing point'),
                        # Line2D([0], [0], marker = '^', color = 'w', markerfacecolor='green', label = 'Rise'),
                        # Line2D([0], [0], marker = 'v', color = 'w', markerfacecolor='blue', label = 'Fall'),
                        # Line2D([0], [0], marker = 'X', color = 'w', markerfacecolor='red', label = 'Other')
                        
                        
                        ]    
    markers =               {0:"s", 1:"s"}
    blackwhite_palette =    ["white", "black"] # 
    marker_sizes = [0.5,0.5]
    if eye_type == "WRITE":
        for j in range(0, physical_CS):
            blackwhite_palette.append(full_training_palette[j])
            markers[2+j]= "X"
            marker_sizes.append(6-j) # 6 for marker size is arbitrary
            legend_label = 'Training point for Phys. CS ' + str(j)
            legend_elements.append(Line2D([0], [0], marker = 'X', color = 'w', markerfacecolor=full_training_palette[j], label = legend_label))

    if eye_type == "READ":
        for j in range(0, logical_CS):
            # Once for rise, another for fall
            blackwhite_palette.append(full_training_palette[j])
            blackwhite_palette.append(full_training_palette[j])
            # Give space in markers so each rank can have rise and fall displayed with that particular color
            # For CS0, a value of 2 will be rise, 3 for fall 
            # For CS1, 4 for rise, 5 for fall
            # Etc.
            markers[2 + 2*j]     = "^"
            markers[2 + 2*j + 1] = "v"
            marker_sizes.append(6-j) 
            marker_sizes.append(6-j)
            legend_label_rise = 'Training point for rise edge rank ' + str(j)
            legend_label_fall = 'Training point for fall edge rank ' + str(j)
            legend_elements.append(Line2D([0], [0], marker = '^', color = 'w', markerfacecolor=full_training_palette[j], label = legend_label_rise))
            legend_elements.append(Line2D([0], [0], marker = 'v', color = 'w', markerfacecolor=full_training_palette[j], label = legend_label_fall))
            
    for i in range(0, bus_width):
            j=0

            detect_pass = 0
            detect_fail = 0
            # For black and white plot, normalize all non-zero fail counts to 1
            while j < len(eyeDiagrams[i*2].eyeData):    
                if (eyeDiagrams[i*2].eyeData[j].fail_cnt > 0 ) or (eyeDiagrams[i*2 +1].eyeData[j].fail_cnt > 0):
                    eyeDiagrams[i*2].eyeData[j].fail_cnt = 1
                    eyeDiagrams[i*2 +1].eyeData[j].fail_cnt = 1
                    detect_fail = 1
                else :
                    detect_pass = 1
                # if eyeDiagrams[i*2 +1].eyeData[j].fail_cnt > 0:
                #     eyeDiagrams[i*2].eyeData[j].fail_cnt = 1
                #     eyeDiagrams[i*2 +1].eyeData[j].fail_cnt = 
                j=j+1
                
            if(detect_fail == 0) :
                print("No failures detected, log may not accurately depict eye!")
                print("Setting an arbitrary point to a failing value to color plot correctly!")
                eyeDiagrams[i*2].eyeData[0].fail_cnt = 1
                eyeDiagrams[i*2 +1].eyeData[0].fail_cnt = 1
            if(detect_pass == 0) :
                print("No passing values detected, log may not accurately depict eye!")
                print("Setting an arbitrary point to a passing value to color plot correctly!")
                eyeDiagrams[i*2].eyeData[0].fail_cnt = 0
                eyeDiagrams[i*2 +1].eyeData[0].fail_cnt = 0
                
            # For black and white plot, training values will be non-zero
            # Shapes of training points changes by plot type
            # To colorize, differs by plot type but color rules remain the same

            # For writes, each DQ will have 1 training point per physical CS
            # Each physical CS will have a different color: Green, Blue, Red, Purple
            if(eye_type == "WRITE"):                
                for j in range(0, physical_CS):            
                    eyeDiagrams[i*2].eyeData.append(eyeDataPoint(2+j,eyeDiagrams[i*2].trainData[j].vref,eyeDiagrams[i*2].trainData[j].delay ))               
                    eyeDiagrams[(i*2) +1].eyeData.append(eyeDataPoint(2+j,eyeDiagrams[i*2 +1].trainData[j].vref,eyeDiagrams[i*2 +1].trainData[j].delay ))

            # For reads, each DQ will have 2 training points per logical CS
            # Each logical CS will have a different color: Green, Blue
            # For rise, shape will be ^
            # For fall, shape will be v
            if(eye_type == "READ"): 
                for j in range(0, logical_CS):       
                    eyeDiagrams[i*2].eyeData.append(eyeDataPoint(2 + 2*j,eyeDiagrams[i*2].trainData[j].vref,eyeDiagrams[i*2].trainData[j].delay ))
                    eyeDiagrams[i*2].eyeData.append(eyeDataPoint(3 + 2*j,eyeDiagrams[i*2 +1].trainData[j].vref,eyeDiagrams[i*2 +1].trainData[j].delay ))

                    eyeDiagrams[(i*2) +1].eyeData.append(eyeDataPoint(2 + 2*j,eyeDiagrams[i*2].trainData[j].vref,eyeDiagrams[i*2].trainData[j].delay ))
                    eyeDiagrams[(i*2) +1].eyeData.append(eyeDataPoint(3 + 2*j,eyeDiagrams[i*2 +1].trainData[j].vref,eyeDiagrams[i*2 +1].trainData[j].delay ))

    for i in range(0, bus_width):
        #vert_pos = int(i/16)*2 + i %2
        #horiz_pos = int(i/2) % 8
        vert_pos = int(i/8)
        horiz_pos = int(i) % 8
       

        

        #print("i = " + str(i) + " Vertical " + str(vert_pos) + ", Horizontal " + str(horiz_pos)) 
        df_rise = pd.DataFrame(eyeDiagrams[i*2].eyeData)
        df_fall = pd.DataFrame(eyeDiagrams[i*2 + 1].eyeData)
        df_sum = df_rise.add(df_fall)
        df_sum = df_sum.div(2)
        eyeVectors_sum.append(df_sum)
    
    saveDiagramToPDF(eyeVectors_sum, blackwhite_palette,markers,marker_sizes, legend_elements, eye_type + " Eye Diagram for DQ bits", eye_type + "_" + filename_no_ext + "_blackwhite", color_plot, pdf_handle )

    
  
   
def saveDiagramToPDF(eyeVectors, color_palette, plot_markers, plot_marker_sizes, legend_elements, plotTitle, imgname, color_plot, pdf_handle ) :
    global max_seen 
    global bus_width
    global eyeDiagrams 
    global subPlotAspectRatio
    once = 0

    for bytelaneDouble in range(0, int(bus_width/16)):
        #TODO: Look into possible refactor of subplot sizing, aspect ratio, spacing
        plt.figure(num=bytelaneDouble, figsize=(2.560, 1.440), dpi=1000, layout='constrained')
        
        # f1, axes1 = plt.subplots(int(bus_width/8),8)
        #f1, axes1 = plt.subplots(figsize=(2,8), layout='constrained')
        f1, axes1 = plt.subplots(2,8)
        
        
        f1.suptitle(plotTitle + "\n" + imgname, fontsize=10)
        #TODO: Smaller second line of plot to show filename
        plt.title(imgname, size='x-small')
        f1.supxlabel("Delay", fontsize=5)
        f1.supylabel("VREF", fontsize=5)
        
        subplot = None
        for i in range(16*bytelaneDouble, 16*bytelaneDouble + 16):
            #Calculate position of each DQ, then plot using Seaborn
            # vert_pos = int(i/8)
            # horiz_pos = int(i) % 8
            vert_pos = int((i-16*bytelaneDouble)/8)
            horiz_pos = int(i) % 8
            if color_plot:
                subplot = sns.scatterplot(data=eyeVectors[i], x="delay", y="vref", hue="fail_cnt", ax=axes1[int(vert_pos),int(horiz_pos)], palette=color_palette, legend=False, edgecolor='None', markers="s", style=1, s=1)
            else:
                subplot = sns.scatterplot(data=eyeVectors[i], x="delay", y="vref", hue="fail_cnt", ax=axes1[int(vert_pos),int(horiz_pos)], palette=color_palette, legend=False, edgecolor='None', markers= plot_markers, style="fail_cnt", size="fail_cnt", sizes=plot_marker_sizes)
            #Remove ticks to graph
            axes1[vert_pos,horiz_pos].tick_params(axis='both', which = 'both', bottom = False, left = False, labelsize = 5)
            axes1[vert_pos,horiz_pos].set_xticklabels([])
            axes1[vert_pos,horiz_pos].set_yticklabels([])

            #Normalize aspect ratio size
            axes1[vert_pos,horiz_pos].set_box_aspect(subPlotAspectRatio) 
            
            #Title each DQ bit
            axes1[vert_pos,horiz_pos].set_title("DQ %d" % i, fontsize=5, pad = 3.0 )
            axes1[vert_pos,horiz_pos].set_xlabel('', fontsize=5 )
            axes1[vert_pos,horiz_pos].set_ylabel('', fontsize=5 )

            #Remove black border around each graph
            axes1[vert_pos,horiz_pos].spines['top'].set_visible(False)
            axes1[vert_pos,horiz_pos].spines['right'].set_visible(False)
            axes1[vert_pos,horiz_pos].spines['bottom'].set_visible(False)
            axes1[vert_pos,horiz_pos].spines['left'].set_visible(False)

        plt.subplots_adjust(left = 0.05, right = 0.95, top = 0.88, bottom = 0.15, hspace = 0.001)#, wspace = 0.01)


        # https://matplotlib.org/stable/gallery/text_labels_and_annotations/custom_legends.html
        # Custom legends
        # legend_elements = [ Line2D([0], [0], marker = 's', color = 'w', markerfacecolor='black', label = 'Failing point'),
        #                     Line2D([0], [0], marker = '^', color = 'w', markerfacecolor='green', label = 'Rise'),
        #                     Line2D([0], [0], marker = 'v', color = 'w', markerfacecolor='blue', label = 'Fall'),
        #                     Line2D([0], [0], marker = 'X', color = 'w', markerfacecolor='red', label = 'Other')
                           
                           
        #                    ]
        f1.legend(handles=legend_elements, loc = 'lower center', fontsize = 'xx-small', ncols=3, bbox_to_anchor=(0.5, 0.5))
        #TODO: Colorbar legend for colorized figure when support is added
        # colors2 =  [ (0, "white"), (1/(2*max_seen), "blue"), (1, "red")]
        # colors2 = colors2.colors.Normalize()
        # f1.colorbar(colors2, cmap)
        
        plt.savefig(imgname + ".png", dpi=1000)
        pp.savefig(f1)
        #plt.show()
        plt.close(f1)
        if once ==0:
            img = Image.open(imgname + ".png")
            img.show()
            once =1


def construct_pdf_table(pdf_handle, fullMarginPlot : allEyeDiagrams, filename_no_ext) :


    global summaryTable 
    global writeTable
    global readTable
    x_ticks= 0 
    y_ticks = 0
    # Extract 
    marginVoltage = []
    marginTiming= []
    vCent=[]
    passingPoints = []

    worstVoltage = 999999999
    worstTiming = 999999999
    smallestPoints = 99999999
    eye_type = fullMarginPlot.type
    eyeDiagrams = fullMarginPlot.diagrams
    df_sum = 0
    for i in range(0, bus_width):
        passingPoints.append(0)
       
        df_rise = pd.DataFrame(eyeDiagrams[i*2].eyeData)
        df_fall = pd.DataFrame(eyeDiagrams[i*2 + 1].eyeData)
        df_sum = df_rise.add(df_fall)
        df_sum = df_sum.div(2)
        #eyeVectors.append(df_sum)
        #print(df_sum)
        if debug:
            print(df_sum.delay.unique())
            print(len(df_sum.delay.unique()))
            print(df_sum.vref.unique())
            print(len(df_sum.vref.unique()))
        x_ticks = len(df_sum.delay.unique())
        y_ticks = len(df_sum.vref.unique())


        #find_margins()
        #print("Number of x ticks/ delay steps = " + str(x_ticks))
        #print("Number of y ticks/ voltage steps = " + str(y_ticks))

        y_start = -1
        y_start_tick= 0
        y_longest = 0
        y_longest_ytick = 0
        y_longest_xtick = 0
        y_prev_valid = 0

        # Find largest y axis opening  
        for x_delay in range(0, x_ticks):
            y_start = -1
            y_prev_valid = 0
            y_start_temp = 0
            for y_vref in range(0, y_ticks):
                #print (y_vref*x_ticks + x_delay)
                if (df_sum.loc[y_vref*x_ticks + x_delay].fail_cnt > 0):
                    if(y_prev_valid == 0):
                        continue
                    else: # if(y_prev_valid == 1):
                        y_prev_valid = 0
                        y_longest_temp = df_sum.loc[(y_vref-1)*x_ticks + x_delay].vref - y_start
                        if(y_longest_temp > y_longest):
                            y_longest= y_longest_temp
                            y_longest_ytick = y_vref -1
                            y_longest_xtick = x_delay
                            y_start_tick = y_start_temp
                else: # Passing Point
                    passingPoints[i] += 1
                    if(y_prev_valid == 0):
                        y_prev_valid = 1
                        y_start = df_sum.loc[y_vref*x_ticks + x_delay].vref
                        y_start_temp = y_vref
                    else: #if(y_prev_valid == 1)
                        continue
            if(y_prev_valid == 1):
                y_prev_valid = 0
                y_longest_temp = df_sum.loc[(y_ticks-1)*x_ticks + x_delay].vref - y_start
                if(y_longest_temp > y_longest):
                    y_longest= y_longest_temp
                    y_longest_ytick = y_ticks-1
                    y_longest_xtick = x_delay
                    y_start_tick = y_start_temp
        if debug:
            print("Voltage For DQ" + str(i))
            print("Longest passing band length = " + str(y_longest))
            print("Band at x tick = " + str(y_longest_xtick))
            print(str(y_start_tick))
            print(str(y_longest_ytick))
        marginVoltage.append(y_longest)

        if(y_longest< worstVoltage):
            worstVoltage = y_longest
        if(passingPoints[i] < smallestPoints):
            smallestPoints = passingPoints[i]

        
        x_start = -1
        x_start_tick= 0
        x_longest = 0
        x_longest_ytick = 0
        x_longest_xtick = 0
        x_prev_valid = 0
        # Find largest x axis opening
        for y_vref in range(0, y_ticks):
            x_start = -1
            x_prev_valid = 0
            
            for x_delay in range(0, x_ticks):
                if (df_sum.loc[y_vref*x_ticks + x_delay].fail_cnt > 0):
                    if(x_prev_valid == 0):
                        continue
                    else: # if(x_prev_valid == 1):
                        x_prev_valid = 0
                        x_longest_temp = df_sum.loc[y_vref*x_ticks + (x_delay - 1)].delay - x_start
                        if(x_longest_temp > x_longest):
                            x_longest= x_longest_temp
                            x_longest_ytick = y_vref
                            x_longest_xtick = x_delay -1
                            x_longest_vref = df_sum.loc[y_vref*x_ticks + (x_delay - 1)].vref
                else: # Passing Point
                    if(x_prev_valid == 0):
                        x_prev_valid = 1
                        x_start = df_sum.loc[y_vref*x_ticks + x_delay].delay
                        x_start_tick = x_delay
                    else: #if(x_prev_valid == 1)
                        continue
            if(x_prev_valid == 1):
                x_prev_valid = 0
                x_longest_temp = df_sum.loc[y_vref*x_ticks + (x_ticks - 1)].delay - x_start
                if(x_longest_temp > x_longest):
                    x_longest= x_longest_temp
                    x_longest_ytick = y_vref
                    x_longest_xtick = x_ticks -1
                    x_longest_vref = df_sum.loc[y_vref*x_ticks + (x_ticks - 1)].vref
        if debug:
            print("Timing For DQ " + str(i))
            print("Longest passing band length = " + str(x_longest))
            print("Band at y tick = " + str(x_longest_ytick))
            print(str(x_start_tick))
            print(str(x_longest_ytick))
        marginTiming.append(x_longest)
        vCent.append( x_longest_vref  ) 
        
        if(x_longest< worstTiming):
            worstTiming = x_longest


    summaryTable.timeStep = str(convertDeltaDelaytoUnitInterval(df_sum.loc[1].delay - df_sum.loc[0].delay))
    summaryTable.voltageStep = str(convertAnyVREFtoVoltage(df_sum.loc[x_ticks].vref - df_sum.loc[0].vref, eye_type))

    # Organize information into structures
    if eye_type == "READ":
        readTable.worstTimingWidth = worstTiming
        readTable.worstVoltageHeight = worstVoltage
        readTable.smallestPassingPoints = smallestPoints
        for i in range(0, bus_width):
            readTable.maxTimingWidth.append(marginTiming[i])
            readTable.maxVoltageHeight.append(marginVoltage[i])
            readTable.totalPassingPointsDQ.append(passingPoints[i])
            #readTable.vCentDQ.append(vCent[i])

    if eye_type == "WRITE":
        writeTable.worstTimingWidth = worstTiming
        writeTable.worstVoltageHeight = worstVoltage
        writeTable.smallestPassingPoints = smallestPoints
        for i in range(0, bus_width):
            writeTable.maxTimingWidth.append(marginTiming[i])
            writeTable.maxVoltageHeight.append(marginVoltage[i])
            writeTable.totalPassingPointsDQ.append(passingPoints[i])
            #writeTable.vCentDQ.append(vCent[i])

    # Create a table
    


    

    tables_fig2 = plt.figure(figsize=(2.560, 1.440), dpi=1000)
    f2, tables_ax2 = plt.subplots(1,1)
    nx = f2.get_figheight()
    ny = f2.get_figwidth()
    f2.set_figheight(ny)
    f2.set_figwidth(nx)
    ncols = 3
    nrows = 29
    tables_ax2.set_xlim(0, 6)
    tables_ax2.set_ylim(0, nrows)
    tables_ax2.set_axis_off()


    overview_columns = ["Pad", "DRAM", "Feature", "Value", "Pad", "DRAM"]
    overview_rows = ["Tool Version", "Voltage Precision", "Unit Interval Precision", "Samples per pass/fail point", "Ranks + Channels", 
                    "Worst " + '\N{GREEK CAPITAL LETTER DELTA}'+ " mV", "Worst " + '\N{GREEK CAPITAL LETTER DELTA}'+ " UI", "Smallest passing points"]
    overview_contents = [summaryTable.tool_version, 
                        str(summaryTable.voltageStep) + " mV", 
                        str(summaryTable.timeStep)  + " UI", 
                        str(summaryTable.num_samples), 
                        str(summaryTable.ranks) + " Ranks, " + str(int(summaryTable.channels)) + " Channels",              
                        str(convertAnyVREFtoVoltage(worstVoltage, eye_type)) + " mV",
                        str(convertDeltaDelaytoUnitInterval(worstTiming))  + " UI", 
                        str(smallestPoints) ]
    
    #Populate overview table
    tables_ax2.annotate(
        xy=(0, nrows),
        text= "Overview", fontsize='9',
        weight='bold',
        ha='center'
    ) 
    for y in range(0, 8):     
        tables_ax2.annotate(
            xy=(2, nrows -1 -y ),
            text= overview_rows[y], fontsize='4',
            weight='bold',
            ha='right'
        )
        tables_ax2.annotate(
            xy=(3, nrows-1 -y ),
            text= overview_contents[y], fontsize='4',
            ha='center'
        )

    #Populate bit mapping table
    tables_ax2.annotate(
        xy=(0, nrows-9 ),
        text= "Bit Mapping", fontsize='7',
        weight='bold',
        ha='center'
    ) 
    tables_ax2.annotate(
        xy=(1, nrows-10 ),
        text= "SoC bit", fontsize='4',
        weight='bold',
        ha='center'
    ) 
    tables_ax2.annotate(
        xy=(0, nrows-10 ),
        text= "DRAM bit", fontsize='4',
        weight='bold',
        ha='center'
    ) 
    if(bus_width>16):
        tables_ax2.annotate(
            xy=(5, nrows-10 ),
            text= "SoC bit", fontsize='4',
            weight='bold',
            ha='center'
        ) 
        tables_ax2.annotate(
            xy=(4, nrows-10 ),
            text= "DRAM bit", fontsize='4',
            weight='bold',
            ha='center'
        )     
    #Create strings for contents of table depending on bit in question
    for i in range(0, bus_width + int(bus_width/8)):
        if(i < bus_width):
            pad_string = "DRAM_DQ" + str(i)
        else:
            pad_string = "DRAM_DM" + str(i - bus_width)

        if(summaryTable.swizzleList[i] >= bus_width):
            dram_string = "PAD_DM" + str(summaryTable.swizzleList[i]-bus_width)  
        else:
            dram_string = "PAD_DQ" + str(summaryTable.swizzleList[i]) 

        tables_ax2.annotate(
            xy=(0 + 4*int(i/18), (nrows -11)-int(i%18) ),
            text= pad_string, fontsize='4',
            ha='center'
        )
        tables_ax2.annotate(
            xy=(1 + 4*int(i/18), (nrows -11)-int(i%18) ),
            text= dram_string, fontsize='4', 
            ha='center'
        )




    pp.savefig(f2)

    columns = ["DQ", "Best\n" + '\N{GREEK CAPITAL LETTER DELTA}'+ " mV", "Best\n" + '\N{GREEK CAPITAL LETTER DELTA}' + " UI" , "Total\nPassing\nPoints" ]
    units = ["", "mV", "UI", "", "mV"]
    nrows = bus_width
    # cell_text = []    
    # for i in range(0, bus_width):
    #     cell_text.append([ 'DQ' + str(i), marginVoltage[i], marginTiming[i]   ]);
    tables_fig = plt.figure(figsize=(2.560, 1.440), dpi=1000)
    f1, tables_ax = plt.subplots(1,1)
    nx = f1.get_figheight()
    ny = f1.get_figwidth()
    f1.set_figheight(ny)
    f1.set_figwidth(nx)
    ncols = 3
    tables_ax.set_xlim(0, 4)
    tables_ax.set_ylim(0, nrows)
    tables_ax.set_axis_off() 
    
    # f1.patch.set_visible(False)
    # tables_ax.axis('off')
    # tables_ax.axis('tight')
    # f1.tight_layout()
    # tables_ax.table(cellText=cell_text, colLabels=columns, loc = 'center', fontsize = 'xx-small')
    

    for i in range(0,1):
        tables_ax.set_xlim(0, 4)
        tables_ax.set_ylim(0, nrows)
        tables_ax.set_axis_off()


        for y in range(0, nrows):
            tables_ax.annotate(
                xy=(0, (nrows - 1)-y ),
                text= "DQ" + str(y), fontsize='4',
                ha='center'
            )
            tables_ax.annotate(
                xy=(1, (nrows - 1)-y ),
                text= str(convertAnyVREFtoVoltage(marginVoltage[y], eye_type)) + " mV", fontsize='4', #"XX.XX V",
                ha='center'
            )
            tables_ax.annotate(
                xy=(2, (nrows - 1)-y),
                text= str(convertDeltaDelaytoUnitInterval(marginTiming[y])) + " UI", fontsize='4', #"XX.XX UI",
                ha='center'
            )
            tables_ax.annotate(
                xy=(3, (nrows - 1)-y),
                text= str(passingPoints[y]), fontsize='4', #"XX.XX UI",
                ha='center'
            )
            # tables_ax.annotate(
            #     xy=(4, (nrows - 1)-y),
            #     text= str(convertAnyVREFtoVoltage(vCent[y], eye_type)) + " mV", fontsize='4', #"XX.XX UI",
            #     ha='center'
            # )
        # Add in table header
        tables_ax.annotate(
            xy=(0, nrows),
            text=columns[0], fontsize = '6',
            weight='bold',
            ha='center'
        )
        tables_ax.annotate(
            xy=(1, nrows),
            text=columns[1], fontsize = '6',
            weight='bold',
            ha='center'
        )
        tables_ax.annotate(
            xy=(2, nrows),
            text=columns[2], fontsize = '6',
            weight='bold',
            ha='center'
        )
        tables_ax.annotate(
            xy=(3, nrows),
            text=columns[3], fontsize = '6',
            weight='bold',
            ha='center'
        )       
        # tables_ax.annotate(
        #     xy=(4, nrows),
        #     text=columns[4], fontsize = '6',
        #     weight='bold',
        #     ha='center'
        # )
    pp.savefig(f1)

# --------------------------------------------------------
# Main function
# --------------------------------------------------------
if __name__ == "__main__":

    # --------------------------------------------------------
    # Global variables
    # --------------------------------------------------------
    global debug
    global vref_limit
    global rddly_limit
    global wrdly_limit
    global eye_type
    global readMarginPlot
    global writeMarginPlot

    # --------------------------------------------------------
    # Parse arguments
    # --------------------------------------------------------
    #TODO: Prepend date
    parser = argparse.ArgumentParser(description='Parse given logfile and output eye data')
    parser.add_argument('logfile',
                        help='Logfile from TEYE tool run')

    # parser.add_argument('--display_rise_fall', action='store_true',
    #                     help='Output the rise and fall eyes used to generate the composite read eye')
    parser.add_argument('--debug', action='store_true',
                        help='Add debug messages', default = None)
    parser.add_argument('--vref_max', action='store',
                        help='Limit the height of vref points processed', default = None)
    parser.add_argument('--rddly_max', action='store',
                        help='Limit the width of rddly points processed (right bound)', default = None)
    parser.add_argument('--wrdly_max', action='store',
                        help='Limit the width of wrdly points processed (right bound)', default = None)
    parser.add_argument('--wrdly_min', action='store',
                        help='Limit the width of wrdly points processed (left bound)', default = None)
    # parser.add_argument('--color', action='store_true',
    #                     help='Display graphs in color, will not display training values')
    # parser.add_argument('--file_prefix_override',
    #                     help='Override prefix in config file')
    #parser.add_argument('--prepend_date', action='store_true', 
    #                     help='Option to prepend date to output data')

    args = parser.parse_args()
    vref_limit = args.vref_max
    rddly_limit = args.rddly_max
    wrdly_limit = args.wrdly_max
    wrdly_min = args.wrdly_min
    debug= args.debug
    color_plot = False #args.color

    # Get the filename isolated from the path it's contained in
    p = Path(__file__).with_name(args.logfile)
    filename = p.absolute()
    filename_no_ext = p.with_suffix('').name

    # --------------------------------------------------------
    # Parse the logfile to fill in data structures
    # readMarginPlot = 2*N diagrams, where N= number of bits
    #   [Bit 0 rise diagram, Bit 0 fall diagram, Bit 1 rise, Bit 1 fall... Bit <N-1> fall diagram]
    #   For the read diagrams, rise/fall will have unique training values for a given bit
    # writeMarginPlot
    #   [Bit 0 rise, Bit 0 fall, ...]
    #   For the write diagrams, rise/fall will have the same training values for a given bit
    #
    # --------------------------------------------------------
    parseLogfile(filename)
    print("Finished parsing logfile!")


    pp = PdfPages(filename_no_ext + ".pdf")

    if "READ" in eye_type:
        construct_pdf_report(pp, readMarginPlot, filename_no_ext, color_plot)
        construct_pdf_table(pp,  readMarginPlot, filename_no_ext)
        
        #displayEyeDiagrams(readMarginPlot, args.display_rise_fall, filename_no_ext, color_plot, pp ) 
    if "WRITE" in eye_type:
        construct_pdf_report(pp, writeMarginPlot, filename_no_ext, color_plot)
        construct_pdf_table(pp,  writeMarginPlot, filename_no_ext)

       # displayEyeDiagrams(writeMarginPlot, args.display_rise_fall, filename_no_ext, color_plot, pp  ) 


    # Close pdf
    pp.close()

    # TODO: Investigate bookmarks into PDF
    #https://www.e-iceblue.com/Tutorials/Python/Spire.PDF-for-Python/Program-Guide/Bookmark/Python-Add-Edit-or-Delete-bookmarks-in-PDF.html

